/*-------------------------------------------------------------------------
 *
 * tupleconn.c
 *	  Routines for tuple connecting.
 *
 * Based on tuplestore.c, tuplesort.c.
 * (c) by Evgen Potemkin evgen_t@mail.ru, 11.2002
 * 
 *-------------------------------------------------------------------------
 */

#include <stdlib.h>
#include <limits.h>
#include "postgres.h"

#include "access/heapam.h"
#include "storage/buffile.h"
#include "utils/tupleconn.h"
#include "utils/tuplesort.h"

#include "access/hash.h"
#include "access/nbtree.h"
//#include "catalog/catname.h"
#include "catalog/pg_amop.h"
#include "catalog/pg_amproc.h"
#include "catalog/pg_operator.h"
#include "miscadmin.h"
#include "utils/datum.h"
#include "utils/catcache.h"
#include "utils/fmgroids.h"
#include "utils/lsyscache.h"
#include "utils/syscache.h"
#include "utils/memutils.h"

#include "parser/parse_expr.h"
#include "nodes/execnodes.h"
#include "executor/executor.h"
#include "c.h"
/*
 * Possible states of a Tupleconn object.	These denote the states that
 * persist between calls of Tupleconn routines.
 */
typedef enum
{
	TCS_INITIAL,				/* Loading tuples; still within memory
								 * limit */
	TCS_WRITEFILE,				/* Loading tuples; writing to temp file */
	TCS_READMEM,				/* Reading tuples; entirely in memory */
	TCS_READFILE				/* Reading tuples from temp file */
} TupConnStatus;


/* 
 * connection tree 
 */
typedef struct _ConnectTree ConnectTree;
struct _ConnectTree{
  ConnectTree *next; /* next sibling*/
  ConnectTree *prev; /* previous sibling */
  ConnectTree *sub;  /* child */
  ConnectTree *sup;  /* parent */
  int  level; /* deep of this node, starts from 1*/
  int  tuple; /* index of tuple in memtuples/used/out_store */
};

/*
 * Information about tuples stored in buffile
 */
typedef struct OutStore{
	int fileno;	/* BufFile fileno */
	long offs;	/* BufFile offs */
	int len;		/* tuple length*/
	HeapTuple tup;	/* in-memory tuple copy, NULL if none */
}OutStore;

/*
 * Datum cache
 */
typedef struct DtCache{
	Datum pntdt_val;	/* cached pntExpr value of tuple*/
	Datum chlddt_val;	/* cached chdlExpr value of tuple*/
	bool pntdt_isnull;	/* cached pntExpr is_null value of tuple*/
	bool chlddt_isnull;	/* cached chdlExpr is_null value of tuple*/
	bool pnt_c;	/* parent value is cached flag*/
	bool chld_c;	/* child value is cached flag*/
} DtCache; 

//cached values of 1 tuple 
typedef struct _HashTableListElem HashTableListElem;
struct _HashTableListElem{
	uint32 pnt_hash;
	bool pnt_isnull;
	Datum pnt_val;
	Datum chld_val;
	uint tup_idx;
	HashTableListElem *next;
};
/*
 * Private state of a Tupleconn operation.
 */
struct Tupleconnstate
{
	TupConnStatus status;		/* enumerated value as shown above */
	long		availMem;		/* remaining memory available, in bytes */
	BufFile    *myfile;			/* underlying file, or NULL if none */

	/*
	 * Tree and supporting stuff, explanation at _performconn function 
	 */
	MemoryContext tree_ctx;	
  ConnectTree *head; /* head of the result list */
	ConnectTree **pnt;  /* parent nodes on curent level*/
	int pntlast;	/* actual amount of parents on current level */
	int pntsize;	/* size of pnt array */
	ConnectTree **chld; /* array of child nodes */
	int chldlast; /* actual amount of childs on current level */
	int chldsize;	/* size of chld array */
	ConnectTree *last; /* last added/fetched node */	
  bool skip_node; /* used in gettuple, mean don't fall to ->sub because 
	                * we're going from that*/
	char *used;	/* array of flags - was tuple (connected already|is null) or not */
  TupleDesc tupDesc;
	AttrNumber level; /* tuple's '_level_' attribute number */
	FmgrInfo connFn; /* comparation function */
	
	OutStore *out_store; /* array of info about tuples in buffile for faster random access */
	DtCache *cache; /* cache for datums returned by ExecEvalExpr */

	HashTableListElem *hash_list;
	HashTableListElem **hash_tbl;
	int hash_tbl_size;
	int nbuckets;
  bool use_hash;  

	void	   *(*copytup) (Tupleconnstate *state, void *tup);
	void		(*writetup) (Tupleconnstate *state, void *tup, bool free, int idx);
	void	   *(*readtup) (Tupleconnstate *state, int idx);

	void	  **memtuples;		/* array of pointers to palloc'd tuples */
	int			memtupsize;		/* allocated length of memtuples array */

	int			tupcount;	/* number of tuples currently present */

	bool		eof_reached;	/* reached EOF (needed for cursors) */

	/* markpos_xxx holds marked position for mark and restore */
	bool		markpos_eof;	/* saved "eof_reached" */
	ConnectTree *markpos_last; /* saved "last" pointer */
	bool		markpos_skip;	/* saved "skip" flag */
  List *priorsList; 
  void *prev_tuple;
};


#define COPYTUP(state,tup)	((*(state)->copytup) (state, tup))
#define WRITETUP(state,tup,free,idx) ((*(state)->writetup) (state, tup, free, idx))
#define READTUP(state,idx)	((*(state)->readtup) (state, idx))
#define LACKMEM(state)		((state)->availMem < 0)
#define USEMEM(state,amt)	((state)->availMem -= (amt))
#define FREEMEM(state,amt)	((state)->availMem += (amt))
#define INIT_GUESS 100000


static Tupleconnstate *tupleconn_begin_common(int maxKBytes);
static void dumptuples(Tupleconnstate *state);

static void *copytup_heap(Tupleconnstate *state, void *tup);
static void writetup_heap(Tupleconnstate *state, void *tup, bool free, int idx);
static void *readtup_heap(Tupleconnstate *state, int idx);
void SelectConnFunction(Oid sortOperator,RegProcedure *sortFunction,
				   SortFunctionKind *kind);
void *setlevel(Tupleconnstate * state,int level, void *tuple, bool free_it, void *prev_tuple);

ConnectTree *add_next(Tupleconnstate * state,ConnectTree *tr,int tup,int level);
ConnectTree *add_sub(Tupleconnstate * state,ConnectTree *tr,int tup,int level);

static uint32 hashFunc(Datum key, int typLen, bool byVal);

void tupleconn_performconn_hash(Tupleconnstate *state, ExprState *parentExpr, ExprState *childExpr, 
    ExprContext *econtext);
/*
 * Init all
 */
static Tupleconnstate *
tupleconn_begin_common(int maxKBytes)
{
	Tupleconnstate *state;
	
	state = (Tupleconnstate *) palloc(sizeof(Tupleconnstate));

	MemSet((char *) state, 0, sizeof(Tupleconnstate));

	state->status = TCS_INITIAL;
	state->availMem = maxKBytes * 1024L;
	state->myfile = NULL;
	state->out_store = NULL;
	
	state->tree_ctx = AllocSetContextCreate(CurrentMemoryContext,"Temporary connby storage",8*1024,8*1024,8*1024);

	state->tupcount = 0;
	if (maxKBytes > 0)
		state->memtupsize = INIT_GUESS;		/* initial guess */
	else
		state->memtupsize = 1;	/* won't really need any space */
	state->memtuples = (void **) palloc(state->memtupsize * sizeof(void *));

  state->pntlast=0;
  state->chldlast=0;
	state->pntsize=INIT_GUESS;		/* initial guess */
	state->chldsize=INIT_GUESS;		/* initial guess */
	state->pnt=(ConnectTree **)palloc(state->pntsize * sizeof(void *));
	state->chld=(ConnectTree **)palloc(state->pntsize * sizeof(void *));

  state->used = (char *) palloc(state->memtupsize * sizeof(char));
  state->level=0;
	return state;
}

Tupleconnstate *
tupleconn_begin_heap(int maxKBytes, TupleDesc tupDesc, Oid connOp, List *priorsList, 
  int levelColNum, double rows,bool useHash)
{
  RegProcedure connFunc;
	SortFunctionKind connFnKind;
	double td;
  MemoryContext morig;
	
	Tupleconnstate *state = tupleconn_begin_common(maxKBytes);

	state->copytup = copytup_heap;
	state->writetup = writetup_heap;
	state->readtup = readtup_heap;

  state->tupDesc=tupDesc;
	
	SelectConnFunction(connOp, &connFunc,
						   &connFnKind);
	if(connFnKind != SORTFUNC_CMP){
	  elog(ERROR,"tupleconn: Can't find suitable function for comparison");
	}
	fmgr_info(connFunc,&state->connFn);
  state->priorsList = priorsList;
  state->level = levelColNum - 1;
  if(useHash){
    state->use_hash = true;
  	td = rows/5;
  	if(td > INT_MAX)
	  	td = INT_MAX;
  	state->nbuckets = (int)td;//assume there is 8 tuples in bucket
	  if(state->nbuckets==0)
		  state->nbuckets=1;
    morig = MemoryContextSwitchTo(state->tree_ctx);
  	state->hash_list = (HashTableListElem *)palloc(sizeof(HashTableListElem)*INIT_GUESS);	//initial guess
	  state->hash_tbl = (HashTableListElem **)palloc(sizeof(void *)*state->nbuckets);	//initial guess
   	MemoryContextSwitchTo(morig);
    MemSet(state->hash_list, 0, sizeof(HashTableListElem)*INIT_GUESS);	//initial guess
    MemSet(state->hash_tbl, 0, sizeof(void *)*state->nbuckets);	//initial guess
  
    state->used = (char *) palloc(state->memtupsize * sizeof(char));
    MemSet(state->used, 0, sizeof(char)*state->memtupsize);	//initial guess
  }else
    state->use_hash = false;

	return state;
}
/*
 * tupleconn_end
 *
 *	Release resources and clean up.
 */
void
tupleconn_end(Tupleconnstate *state)
{
	int	i;
	/* common frees */
	pfree(state->pnt);
	pfree(state->chld);
	pfree(state->used);
	/* free tree, if any */
	state->last = state->head;
	// release tree and cache
	MemoryContextDelete(state->tree_ctx);
	/* free tuples, from out_store or memtuples*/
	if (state->myfile){
		BufFileClose(state->myfile);
		for(i=0;i<state->tupcount;i++){
			if(state->out_store[i].tup!=NULL){
				pfree(state->out_store[i].tup);
				FREEMEM(state,state->out_store[i].len);
				state->out_store[i].tup=NULL;
			}
		}
	}else{
		for (i = 0; i < state->tupcount; i++)
			pfree(state->memtuples[i]);
		pfree(state->memtuples);
	}
}

/*
 * Accept one tuple while collecting input data.
 *
 * Note that the input tuple is always copied; the caller need not save it.
 */
void
tupleconn_puttuple(Tupleconnstate *state, void *tuple,int head)
{
	/*
	 * Copy the tuple.	(Must do this even in WRITEFILE case.)
	 */
	tuple = COPYTUP(state, tuple);

	/* common thing */
  if(head){//it's a head tuple,add it to head list
    //add autogrow for pnt
	  state->last = add_next(state,state->last,state->tupcount,1);
    if(!state->head) state->head = state->last;
    state->pnt[state->pntlast++] = state->last;
		state->used[state->tupcount] = 1;
  }else{
	  state->used[state->tupcount] = 0;
	}
	
	switch (state->status)
	{
		case TCS_INITIAL:

			/*
			 * Stash the tuple in the in-memory array.
			 */
			if (state->tupcount >= state->memtupsize-1)
			{
				/* Grow the arrays as needed. */
				state->memtupsize *= 2;
				state->memtuples = (void **)
					repalloc(state->memtuples,
							 state->memtupsize * sizeof(void *));
				state->used = (char *)
					repalloc(state->used,
							 state->memtupsize * sizeof(char));
				MemSet((char *) (state->used+sizeof(char)*(state->memtupsize>>1)), 0, sizeof(char)*(state->memtupsize>>1));
			}
			state->memtuples[state->tupcount++] = tuple;
			/*
			 * Done if we still fit in available memory.
			 */
			if (!LACKMEM(state))
				break;

			/*
			 * Nope; time to switch to tape-based operation.
			 */
			state->myfile = BufFileCreateTemp(false);			
			state->status = TCS_WRITEFILE;
			state->out_store = palloc(state->memtupsize * sizeof(OutStore));
			dumptuples(state);
			break;
		case TCS_WRITEFILE:
			if (state->tupcount >= state->memtupsize-1)
			{
				/* Grow the arrays as needed. */
				state->memtupsize *= 2;
				state->out_store = (OutStore *)
					repalloc(state->out_store,
							 state->memtupsize * sizeof(OutStore));
				state->used = (char *)
					repalloc(state->used,
							 state->memtupsize * sizeof(char));
				MemSet((char *) (state->used+sizeof(char)*(state->memtupsize>>1)), 0, sizeof(char)*(state->memtupsize>>1));
			}
			WRITETUP(state, tuple,!head,state->tupcount++);
			break;
		default:
			elog(ERROR, "tupleconn_puttuple: invalid state (internal error)");
			break;
	}
}

/*
 * All tuples have been provided; finish writing.
 */
void
tupleconn_donestoring(Tupleconnstate *state)
{
	switch (state->status)
	{
		case TCS_INITIAL:
			/*
			 * We were able to accumulate all the tuples within the
			 * allowed amount of memory.  Just set up to connect and scan them.
			 */
			state->status = TCS_READMEM;
			break;
		case TCS_WRITEFILE:
			/*
			 * Set up for connecting/reading from tape.
			 */
			state->status = TCS_READFILE;
			break;
		default:
			elog(ERROR, "tupleconn_donestoring: invalid state ");
			break;
	}
	state->eof_reached = false;
	state->markpos_eof = false;
	state->last=state->head;
}

/*
 * tupleconn_performconn: perform connection on tuples
 * Algorithm: in puttuple has been made list of top parent nodes,
 * in each iteration we try to find all non-connected tuples which
 * 'child' attribute is equal to 'parent' attribute in one of parent 
 * nodes, if so - tuple becomes a child of corresponding parent node.
 * at end of iteration collected childs becomes the parents for next 
 * iteration. 
 * If no childs were find algorithm stops.
 * Scan for childs in one iteration going on full array of stored 
 * tuples - this preserves order of tuples from subplan. for example 
 * if subplan was alphabetically sorted, childs on one level of each 
 * parent will be also alphabetically sorted.
 * In case of file storage at end of algorithm all tuples resides only 
 * on tape.
 * 
 * Clause was:
 * CONNECT BY PRIOR expr Op expr
 * here:
 * PRIOR expr is - parentExpr, checked against parent tuple
 * Op - is connOp operator, performs comparation
 * expr (the rest) - is childExpr, checked against child tuple
 */
void
tupleconn_performconn(Tupleconnstate *state, 
					ExprState *parentExpr, ExprState *childExpr, ExprContext *econtext){
  int ok=0,i,j,ti,level=1;
	ConnectTree **t;
	int32 is_parent;
	Datum dt1,dt2;
	bool p_isnull,isnull,conn;
	HeapTuple pnt,chld;
	TupleDesc tupDesc;
	TupleTableSlot *slot;
	MemoryContext morig;
  int16	pnt_typlen;
	bool	pnt_typbyval;
  int16	chld_typlen;
	bool	chld_typbyval;
	Size	datalen;
	char	*newVal;
  ExprState *parentExprState,*childExprState;

	if(!state->head) return; /* trivial case, don't connect anything */
  if(state->use_hash)
    return tupleconn_performconn_hash(state, parentExpr, childExpr, econtext);

	/* check status */
	switch(state->status){
		case TCS_READMEM:
		case TCS_READFILE:
			break;
		default:
			elog(ERROR,"tupleconn: invalid state in performconn (internal error)");
	}

  parentExprState=parentExpr;
  childExprState=childExpr;
	tupDesc=state->tupDesc;
	
	/* get child and parent exprs typlens for cache */	
	get_typlenbyval(exprType(parentExprState->expr), &pnt_typlen, &pnt_typbyval);
	get_typlenbyval(exprType(childExprState->expr), &chld_typlen, &chld_typbyval);
	
	/* alloc cache in temp space */
	morig = MemoryContextSwitchTo(state->tree_ctx);
	state->cache = (DtCache *)palloc(sizeof(DtCache)*state->tupcount);	
	MemoryContextSwitchTo(morig);
	MemSet(state->cache, 0, sizeof(DtCache)*state->tupcount);

	/* make for ExecEvalExpr temporary slot */
        slot= MakeSingleTupleTableSlot(tupDesc);
/*	slot=makeNode(TupleTableSlot);
	slot->tts_tupleDescriptor = tupDesc;
	slot->tts_buffer = InvalidBuffer;
	slot->tts_shouldFreeDesc = false;
	slot->tts_shouldFree = true;
//	slot->tts_descIsNew = false;
	*/
	/* set slot for ExecExvalExpr */
	econtext->ecxt_scantuple = slot;
	
  while(!ok){
	  ok=1;
 		for(i=0;i<state->tupcount;i++){//scan through array of tuples
			/* skip already connected and null tuples */
			CHECK_FOR_INTERRUPTS();
     	if(!state->used[i]){
				ResetExprContext(econtext);
				/* get tuple for connecting */
				/* if cached - use cache, in not - retrieve tuple, and build cache */
				if(state->cache[i].chld_c){
					dt1 = state->cache[i].chlddt_val;
					isnull = state->cache[i].chlddt_isnull;
				}else{
					if(state->status == TCS_READMEM) {
						chld=(HeapTuple)state->memtuples[i];
						slot->tts_tuple = chld;
                                                slot->tts_nvalid= 0;
						dt1 = ExecEvalExpr(childExprState,econtext,&isnull,NULL);
					}	else { //READFILE
						chld=READTUP(state,i);
						Assert(chld!=NULL);
						slot->tts_tuple = chld;
                                                slot->tts_nvalid= 0;
						/* get value of child expr,
						 * in case of variable node is equal to 
					   * dt1=heap_getattr(chld, attno1, tupDesc, &isnull); 
						 */
						dt1 = ExecEvalExpr(childExprState,econtext,&isnull,NULL);						
					}
					/* no need in storing isnull, because when we first time 
					 * evaluating expr and it's null then it marked as used 
					 * and don't checked anymore, thus cache not involved
					 */					 
					state->cache[i].chld_c = true;
					/* store it if it's not null */
					if(!isnull){
						if(chld_typbyval){
							state->cache[i].chlddt_val = dt1;
						}else {
							/* copy datum data to temp space*/
							datalen = datumGetSize(dt1, false, chld_typlen);
							morig = MemoryContextSwitchTo(state->tree_ctx);
							newVal = (char *) palloc(datalen);
							MemoryContextSwitchTo(morig);
							memcpy(newVal, DatumGetPointer(dt1), datalen);
							state->cache[i].chlddt_val = PointerGetDatum(newVal);
							state->cache[i].chlddt_isnull = false;
							dt1 = PointerGetDatum(newVal);
						}
					}
					/* free tuple, since we don't need it further, till _gettuple */
					if(state->status != TCS_READMEM) {
						pfree(chld);
						FREEMEM(state,((HeapTuple)chld)->t_len+HEAPTUPLESIZE);
						state->out_store[i].tup=NULL;						
					}
				}
				conn=false;
				if(!isnull){
					/* scan through nodes array of previous level,
					 * until connect it or array is exhausted
					 */
				  for(j=0;j<state->pntlast && !conn;j++){
						/* get parent tuple */
						/* if cached - use cache, in not - retrieve tuple, and build cache */
						if(state->cache[state->pnt[j]->tuple].pnt_c){
							dt2 = state->cache[state->pnt[j]->tuple].pntdt_val;
							p_isnull = state->cache[state->pnt[j]->tuple].pntdt_isnull;
						}else{
							if(state->status == TCS_READMEM) {
							  pnt=(HeapTuple)state->memtuples[state->pnt[j]->tuple];
								slot->tts_tuple = pnt;
                                                                slot->tts_nvalid= 0;
								dt2 = ExecEvalExpr(parentExprState,econtext,&p_isnull,NULL);
							} else {
								// if tuple parent value is not cached and tuple isn't in mem, 
								// then read it and build cache
								pnt=state->out_store[state->pnt[j]->tuple].tup;
						
								if(pnt==NULL){
									pnt=READTUP(state,state->pnt[j]->tuple);
								}
								Assert(pnt!=NULL); /*elog(ERROR,"tupleconn: parent tuple is null (internal error)");*/

								slot->tts_tuple = pnt;
                                                                slot->tts_nvalid= 0;
								/* get value of parent expr */
								dt2 = ExecEvalExpr(parentExprState,econtext,&p_isnull,NULL);
							}
							state->cache[state->pnt[j]->tuple].pntdt_isnull = p_isnull;
							state->cache[state->pnt[j]->tuple].pnt_c = true;
							/* cache it, if it's not null*/
							if(!p_isnull){
								if(pnt_typbyval){
									state->cache[state->pnt[j]->tuple].pntdt_val = dt2;
								}else{
									/* copy datum data to temp space */
									datalen = datumGetSize(dt2, false, pnt_typlen);
									morig = MemoryContextSwitchTo(state->tree_ctx);
									newVal = (char *) palloc(datalen);
									MemoryContextSwitchTo(morig);
									memcpy(newVal, DatumGetPointer(dt2), datalen);
									state->cache[state->pnt[j]->tuple].pntdt_val = PointerGetDatum(newVal);
									state->cache[state->pnt[j]->tuple].pntdt_isnull = false;
									dt2 = PointerGetDatum(newVal);
								}
							}
							/* free tuple, same as child */
							if(state->status != TCS_READMEM) {
								pfree(pnt);
								FREEMEM(state,((HeapTuple)pnt)->t_len+HEAPTUPLESIZE);
								state->out_store[state->pnt[j]->tuple].tup = NULL;
							}
						}
						if(!p_isnull){
							/* apply connOp operator on parent and child expr results,
							 * if 0 (in case of CMP means they're equal) connect them
							 */
							is_parent=DatumGetInt32(FunctionCall2(&state->connFn,dt1,dt2));
							if(is_parent==0){
		        	  ok=0;
								/* stop scan of parents */
								conn=true;
								/* connect tuples (make node of the connect tree)*/
 			        	state->chld[state->chldlast++]=add_sub(state,state->pnt[j],i,level+1);
								state->used[i]=1;
								if(state->chldlast>=state->chldsize-1){
									/* grow array of connected tuples as necessary */
									state->chldsize *= 2;
									state->chld = (ConnectTree **)
											repalloc(state->chld,
									 		state->chldsize * sizeof(void *));
								}
  	       		}
						}
 		    	}
				}else{
					/* mark it as used since it has null child value and can't be 
					 * connected to any node
					 * may be better to add nullcheck on this field at parse stage to whereClause, 
					 */
					state->used[i]=1;
				}
			}
 	  }
		/* swap pnt & chld arrays */
		t=state->pnt;
		ti=state->pntsize;

   	state->pnt=state->chld;
		state->pntsize=state->chldsize;
		state->pntlast=state->chldlast;

		state->chld=t;
		state->chldsize=ti;
		state->chldlast=0;

 	  level++;
 	}
	/* free anything temporal */
	ResetExprContext(econtext);
	//pfree(slot);
        ExecDropSingleTupleTableSlot(slot);
	econtext->ecxt_scantuple = NULL;
	pfree(state->cache);
	state->cache=NULL;
}

void
tupleconn_performconn_hash(Tupleconnstate *state,
					ExprState *parentExpr, ExprState *childExpr, ExprContext *econtext){
  int ok=0,i,j,ti,level=1,k;
	ConnectTree **t;
	int32 is_parent;
	Datum dt1,dt2;
	bool isnull,conn;
	HeapTuple tuple;
	TupleDesc tupDesc;
	TupleTableSlot *slot;
	MemoryContext morig;
  int16	pnt_typlen;
	bool	pnt_typbyval;
  int16	chld_typlen;
	bool	chld_typbyval;
	Size	datalen;
	char	*newVal;
	uint32 h1,h2;
	uint bucket;  
	HashTableListElem *htle;
  ExprState *parentExprState,*childExprState;

	if(!state->head) return; /* trivial case, don't connect anything */
	/* check status */
	switch(state->status){
		case TCS_READMEM:
		case TCS_READFILE:
			break;
		default:
			elog(ERROR,"tupleconn: invalid state in performconn (internal error)");
	}

  parentExprState=parentExpr;
  childExprState=childExpr;

	tupDesc=state->tupDesc;
	/* get child and parent exprs typlens for cache */	
	get_typlenbyval(exprType(parentExprState->expr), &pnt_typlen, &pnt_typbyval);
	get_typlenbyval(exprType(childExprState->expr), &chld_typlen, &chld_typbyval);
	
	/* alloc cache in temp space */
	morig = MemoryContextSwitchTo(state->tree_ctx);
	state->hash_list = (HashTableListElem *)palloc(sizeof(HashTableListElem)*state->tupcount);	
	MemoryContextSwitchTo(morig);
	MemSet(state->hash_list, 0, sizeof(HashTableListElem)*state->tupcount);

	/* make for ExecEvalExpr temporary slot */
        slot= MakeSingleTupleTableSlot(tupDesc);
        /*
	slot=makeNode(TupleTableSlot);
	slot->tts_tupleDescriptor = tupDesc;
	slot->tts_buffer = InvalidBuffer;
	slot->tts_shouldFreeDesc = false;
	slot->tts_shouldFree = true;
//	slot->ttc_descIsNew = false;
	*/
	/* set slot for ExecExvalExpr */
	econtext->ecxt_scantuple = slot;

  /*
   * build hash table
   * for all tuples, calculate hash for childExpr. if tuple is head then 
   * calc parentExpr hash instead. if either Expr is null, throw tuple out.
   */
  for(i=0;i<state->tupcount;i++){
    //get tuple
    if(state->status == TCS_READMEM) {
			tuple=(HeapTuple)state->memtuples[i];
		}	else { //READFILE
			tuple=READTUP(state,i);
			Assert(tuple!=NULL);
  	}
    
		slot->tts_tuple = tuple;
                slot->tts_nvalid= 0; //invalidate cache
    state->hash_list[i].tup_idx = i;
    ResetExprContext(econtext);
		if(state->used[i]){ //for now it means 'head'
  	  dt2 = ExecEvalExpr(parentExprState,econtext,&isnull,NULL);
      if(!isnull) {
        h2 = hashFunc(dt2, pnt_typlen, pnt_typbyval);
  			//fill pnt_hash value for perfconn
				//no need to add this tuple to hash_table, we never will use it as child tuple
	      state->hash_list[i].pnt_hash = h2;

				if(pnt_typbyval || state->status == TCS_READMEM){
					state->hash_list[i].pnt_val = dt2;
				}else {
					/* copy datum data to temp space*/
					datalen = datumGetSize(dt2, false, pnt_typlen);
					morig = MemoryContextSwitchTo(state->tree_ctx);
					newVal = (char *) palloc(datalen);
					MemoryContextSwitchTo(morig);
					memcpy(newVal, DatumGetPointer(dt2), datalen);
					state->hash_list[i].pnt_val = PointerGetDatum(newVal);
					state->hash_list[i].pnt_isnull = false;
				}
				/* free tuple, since we don't need it further, till _gettuple */
      }else{
        state->hash_list[i].pnt_isnull = true;
      }
		}else{
			uint bucket;
			HashTableListElem *htle;
			MemoryContext morig;

  	  dt1 = ExecEvalExpr(childExprState,econtext,&isnull,NULL);
			if(!isnull) {// sanity check
        
        h1 =	hashFunc(dt1, chld_typlen, chld_typbyval);
  			bucket = h1 % state->nbuckets;

				if(chld_typbyval || state->status == TCS_READMEM){
					state->hash_list[i].chld_val = dt1;
				}else {
						/* copy datum data to temp space*/
					datalen = datumGetSize(dt1, false, chld_typlen);
					morig = MemoryContextSwitchTo(state->tree_ctx);
					newVal = (char *) palloc(datalen);
					MemoryContextSwitchTo(morig);
					memcpy(newVal, DatumGetPointer(dt1), datalen);
					state->hash_list[i].chld_val = PointerGetDatum(newVal);
				}
        // add tuple to bucket
				if(state->hash_tbl[bucket]){
					for(k=0,htle = state->hash_tbl[bucket];htle->next;htle=htle->next);//k++
					htle->next=&state->hash_list[i];
				}else{
					state->hash_tbl[bucket] = &state->hash_list[i];
				} 
      }else{
        //it's null, throw it out
        state->used[i] = 1;
      }
	
  	}
		/* free tuple, since we don't need it further, till _gettuple */
  	if(state->status != TCS_READMEM) {
			pfree(tuple);
			FREEMEM(state,((HeapTuple)tuple)->t_len+HEAPTUPLESIZE);
			state->out_store[i].tup=NULL;						
		}
  }
  /*
   * perform connection of tuples
   * for each head tuple calculate bucket from parent hash, then check connect conditions on 
   * every tuple in bucket list, if accepted add to tree.
   */
  /*
   * process list of parents, if no childs are found (ok==1), then tree building is complete.
   */
  while(!ok){
	  ok=1;
		CHECK_FOR_INTERRUPTS();
		conn=false;
          
		for(j=0;j<state->pntlast;j++){
			int pnt_tup_idx;
			/* get parent tuple */
			/* if cached - use cache, in not - retrieve tuple, and build cache */
			pnt_tup_idx = state->pnt[j]->tuple;
			if(state->hash_list[pnt_tup_idx].pnt_isnull) continue;
			h2 = state->hash_list[pnt_tup_idx].pnt_hash;
			dt2 = state->hash_list[pnt_tup_idx].pnt_val;

			/* get bucket */
			bucket = h2 % state->nbuckets;

      i=0;k=0;
      /* search bucket for child tuples */
			for(htle=state->hash_tbl[bucket];htle;htle = htle->next){
				if(!htle)
					break;
        //skip it if null or the self
				if(state->used[htle->tup_idx] || pnt_tup_idx == htle->tup_idx )
					continue;
        //check connect condition
				is_parent=DatumGetInt32(FunctionCall2(&state->connFn,htle->chld_val,dt2));
				if(is_parent !=0)//isn't accepted
					continue;
         // found one or more childs
		    ok=0;
        // calc parentExpr hash, it will be used on next round
				if(!htle->pnt_hash){
					Datum dt;
					bool isnull;
					ResetExprContext(econtext);
          
          if(state->status == TCS_READMEM) {
      			tuple=(HeapTuple)state->memtuples[htle->tup_idx];
      		}	else { //READFILE
      			tuple=READTUP(state,htle->tup_idx);
      			Assert(tuple!=NULL);
        	}
					slot->tts_tuple = tuple;
                                        slot->tts_nvalid = 0;

				  dt = ExecEvalExpr(parentExprState,econtext,&isnull,NULL);
			    if(!isnull) {
            int tttt;
						htle->pnt_hash = hashFunc(dt, pnt_typlen, pnt_typbyval);
            tttt = htle->pnt_hash % state->nbuckets;
      			if(pnt_typbyval){
      				htle->pnt_val = dt;
      			}else {
      				/* copy datum data to temp space*/
      				datalen = datumGetSize(dt, false, pnt_typlen);
      				morig = MemoryContextSwitchTo(state->tree_ctx);
      				newVal = (char *) palloc(datalen);
      				MemoryContextSwitchTo(morig);
      				memcpy(newVal, DatumGetPointer(dt), datalen);
      				htle->pnt_val = PointerGetDatum(newVal);
      				htle->pnt_isnull = false;
      			}
          	if(state->status != TCS_READMEM) {
        			pfree(tuple);
        			FREEMEM(state,((HeapTuple)tuple)->t_len+HEAPTUPLESIZE);
        			state->out_store[i].tup=NULL;						
        		}
					}else
						htle->pnt_isnull = true;
				}
				/* connect tuples (make node of the connect tree)*/
 		  	state->chld[state->chldlast++]=add_sub(state,state->pnt[j],htle->tup_idx,level+1);
        state->used[htle->tup_idx] = 1;
				if(state->chldlast>=state->chldsize-1){
					/* grow array of connected tuples as necessary */
					state->chldsize *= 2;
					state->chld = (ConnectTree **)
							repalloc(state->chld,
					 		state->chldsize * sizeof(void *));
				}
        i++;
  	 	}
 		}
		/* swap pnt & chld arrays */
		t=state->pnt;
		ti=state->pntsize;

   	state->pnt=state->chld;
		state->pntsize=state->chldsize;
		state->pntlast=state->chldlast;

		state->chld=t;
		state->chldsize=ti;
		state->chldlast=0;
 	  level++;
 	}
	/* free anything temporal */
	ResetExprContext(econtext);
//	pfree(slot);
        ExecDropSingleTupleTableSlot(slot);
	econtext->ecxt_scantuple = NULL;
	state->cache=NULL;
}


/* set _level_ column to proper value 
 * first decompose tuple. if we're found _level_ column,
 * set it to proper value. them form new tuple.
 * in tape case free original tuple.
 */
void *
setlevel(Tupleconnstate *state,int level, void *tuple,bool free_it, void *prev_tuple){
#define REALLOCATTRS 64
	Datum valuesArray[REALLOCATTRS];
	char nullsArray[REALLOCATTRS];
	Datum pvaluesArray[REALLOCATTRS];
	char pnullsArray[REALLOCATTRS];
	Datum *values,*pvalues;
	char *nulls,*pnulls;
	HeapTuple newtup,tup = (HeapTuple)tuple,ptup = (HeapTuple)prev_tuple;
	TupleDesc tdesc = state->tupDesc;
	int natts,i,c;
  ListCell *tl;
 	bool		isnull;

	if(!tuple) return NULL;
	natts = tdesc->natts;
	/* prepare arrays */
	if(natts>REALLOCATTRS){
		values = palloc(natts * sizeof(Datum));
		nulls = palloc(natts * sizeof(char));
    if(prev_tuple){
  		pvalues = palloc(natts * sizeof(Datum));
	  	pnulls = palloc(natts * sizeof(char));
    }else{
  		pvalues = NULL;
	  	pnulls = NULL;
    }
	}else{
		values = valuesArray;
		nulls = nullsArray;
		pvalues = pvaluesArray;
		pnulls = pnullsArray;
	}
	/* decompose tuple and substitute attr _level_ with real value */
	for (i = 0; i < natts; i++){
		if(i != state->level){
			values[i] = heap_getattr(tup,
								 i + 1,
								 tdesc,
								 &isnull);
			if (isnull)
				nulls[i] = 'n';
			else
				nulls[i] = ' ';
		}else{
			values[state->level]=Int32GetDatum(level);
			nulls[i] = ' ';
		}
	}
  if(state->priorsList){
    List *ttl;

    if(prev_tuple)
  	for (i = 0; i < natts; i++){
		  if(i != state->level){
			  pvalues[i] = heap_getattr(ptup,
				  			 i + 1,
								 tdesc,
								 &isnull);
  			if (isnull)
	  			pnulls[i] = 'n';
		  	else
			  	pnulls[i] = ' ';
  		}
	  }

    foreach(tl,state->priorsList){
      ListCell *lc;
      ttl=lfirst(tl);
      lc=list_head(ttl);
      i=lfirst_int(lc);
      c=lfirst_int(lnext(lc));
      if(prev_tuple){
        nulls[i] = pnulls[c];
        values[i] = pvalues[c];
      }else{
        nulls[i] = 'n';
      }
		}
  }
	/* form new */
	tup = heap_formtuple(state->tupDesc,values,nulls);
	if(natts>REALLOCATTRS){
		pfree(values);
		pfree(nulls);
    if(prev_tuple){
  		pfree(pvalues);
	  	pfree(pnulls);
    }
	}
	if(free_it){
		/* make full copy of modified tuple and free original */
		newtup = heap_copytuple(tup);
		tup = newtup;
		FREEMEM(state,((HeapTuple)tuple)->t_len+HEAPTUPLESIZE);
		heap_freetuple(tuple);
		tup = newtup;
    //also free prev_tuple, if any
    if(prev_tuple){
      FREEMEM(state,((HeapTuple)prev_tuple)->t_len+HEAPTUPLESIZE);
      heap_freetuple(prev_tuple);
    }
	}
  return tup;
}
/*
 * Fetch the next tuple in forward only direction.
 * Returns NULL if no more tuples. If should_free is set, the
 * caller must pfree the returned tuple when done with it.
 */
/* FIXME: add backward direction in future. may be.
 */
void *
tupleconn_gettuple(Tupleconnstate *state,
					bool *should_free)
{
	void	   *tup=NULL;
	int 		level=0;
  void    *prev_tuple=NULL;
	  
	if (state->eof_reached || !state->head) return NULL;
	/* check status */
	switch (state->status)
	{
		case TCS_READMEM:
			*should_free = false;
			break;
		case TCS_READFILE:
			*should_free = true;
			break;
		default:
			elog(ERROR, "tupleconn_gettuple: invalid state");
			return NULL;		/* keep compiler happy */
	}

	while(!tup) {
		if(!state->skip_node) {
			if(state->status == TCS_READMEM){
				tup=state->memtuples[state->last->tuple];
        if(state->last->sup){
          prev_tuple = state->memtuples[state->last->sup->tuple];
        }else{
          prev_tuple = NULL;
        }
			}else{
				tup=READTUP(state,state->last->tuple);
        if(state->priorsList && state->last->sup){
          prev_tuple = READTUP(state,state->last->sup->tuple);
          // same as a bit below
          state->out_store[state->last->sup->tuple].tup=NULL;
        }else{
          prev_tuple = NULL;
        }
				/* will be freed in setlevel(), but setlevel()
				 * can't clear tuple's out_store[] cell, so clear it here 
				 */
				state->out_store[state->last->tuple].tup=NULL;
			}
			level=state->last->level;
		}
		if(!state->skip_node && state->last->sub) state->last=state->last->sub;
		else if(state->last->next){
			state->last=state->last->next;
			state->skip_node = false;
		}	else if(state->last->sup){
		  state->last=state->last->sup;
			state->skip_node=true;
		}	else {
		  state->eof_reached = true;
		  break;
		}
	}
  if(!state->priorsList){
    prev_tuple = NULL;
  }
	return setlevel(state,level,tup,*should_free,prev_tuple);
}

/*
 * dumptuples - remove tuples from memory and write to tape
 */
static void
dumptuples(Tupleconnstate *state)
{
	int			i,j;
	bool	b;
	for (i = 0, j = 0; i < state->tupcount; i++){
		/* don't free pnt list, because we will use it soon, in tupleconn_performconn() */
		if(j < state->pntlast && i == state->pnt[j]->tuple){
			b=false;
			j++;
		}else b=true;
		WRITETUP(state, state->memtuples[i],b,i);
	}
}

/*
 * tupleconn_rescan		- rewind and replay the scan
 */
void
tupleconn_rescan(Tupleconnstate *state)
{

	/* check status */
	switch (state->status)
	{
		case TCS_READMEM:
		case TCS_READFILE:
			break;
		default:
			elog(ERROR, "tupleconn_rescan: invalid state");
			break;
	}
	state->eof_reached = false;
	state->markpos_eof = false;
	
	state->last = state->head;
}

/*
 * tupleconn_markpos	- saves current position in the tuple sequence
 */
void
tupleconn_markpos(Tupleconnstate *state)
{

	/* check status */
	switch (state->status)
	{
		case TCS_READMEM:
		case TCS_READFILE:
			break;
		default:
			elog(ERROR, "tupleconn_markpos: invalid state");
			break;
	}
	state->markpos_eof = state->eof_reached;

	/* file/memtuples positions can be retrieved by state->last
	 * so don't save them
	 */
	state->markpos_last = state->last;
	state->markpos_skip = state->skip_node;
}

/*
 * tupleconn_restorepos - restores current position in connection tree to
 *						  last saved position
 */
void
tupleconn_restorepos(Tupleconnstate *state)
{

	/* check status */
	switch (state->status)
	{
		case TCS_READMEM:
		case TCS_READFILE:
			break;
		default:
			elog(ERROR, "tupleconn_restorepos: invalid state");
			break;
	}
	state->eof_reached = state->markpos_eof;

	state->last = state->markpos_last;
	state->skip_node = state->markpos_skip;
}

/*
 * Routines specialized for HeapTuple case
 */

static void *
copytup_heap(Tupleconnstate *state, void *tup)
{
	HeapTuple	tuple = (HeapTuple) tup;

	USEMEM(state, HEAPTUPLESIZE + tuple->t_len);
	return (void *) heap_copytuple(tuple);
}

/*
 * tree building procedures
 */
 
/* add new node next to given, set and return it */
ConnectTree *
add_next(Tupleconnstate *state,ConnectTree *tr,int tup,int level){
  ConnectTree *t;
  MemoryContext morig;
  
	morig=MemoryContextSwitchTo(state->tree_ctx);
  t=palloc(sizeof(ConnectTree));
	MemoryContextSwitchTo(morig);
  memset(t,0,sizeof(ConnectTree));
  t->tuple=tup;
  t->level=level;
  if(tr){
    tr->next=t;
    t->prev=tr;
  }
  return t;
}

/* add new node as child to given, set and return it */
ConnectTree *
add_sub(Tupleconnstate *state,ConnectTree *tr,int tup,int level){
  ConnectTree *t,*t1;
  MemoryContext morig;
  
	morig=MemoryContextSwitchTo(state->tree_ctx);
  t=palloc(sizeof(ConnectTree));
	MemoryContextSwitchTo(morig);
  memset(t,0,sizeof(ConnectTree));
  t->tuple=tup;
  t->level=level;
  if(!tr->sub){
    tr->sub=t;  
    t->sup=tr;
  }else{
    for(t1=tr->sub;t1->next;t1=t1->next);
    t1->next=t;
    t->prev=t1;
    t->sup=tr;
  }
  return t;
}


/*
 * File storage procedures
 * this part taken from tuplestore.c, with addition for 
 * maintaining out_store[].
 * because of all on-tape tuple placement info is stored in 
 * out_store[], don't bother to write it on a tape, only
 * tuple body is stored.
 */
 
/*
 * We don't bother to write the HeapTupleData part of the tuple.
 */

static void
writetup_heap(Tupleconnstate *state, void *tup, bool free, int idx)
{
	HeapTuple	tuple = (HeapTuple) tup;

	/* fill placement info */
	state->out_store[idx].len = tuple->t_len;
	BufFileTell(state->myfile,
				&state->out_store[idx].fileno,
				&state->out_store[idx].offs);
	
	if (BufFileWrite(state->myfile, (void *) tuple->t_data,
					 tuple->t_len) != (size_t) tuple->t_len)
		elog(ERROR, "tupleconn: write failed");
	
	/* explanation in dumptuples */
	if(free){
//		FREEMEM(state, HEAPTUPLESIZE + tuple->t_len);
		heap_freetuple(tuple);
		state->out_store[idx].tup = NULL;		
	}else{
		state->out_store[idx].tup = (HeapTuple)tup;
	}
}

static void *
readtup_heap(Tupleconnstate *state, int idx)
{
	unsigned int tuplen = state->out_store[idx].len + HEAPTUPLESIZE;
	/* add block readings */
	HeapTuple	tuple = (HeapTuple) palloc(tuplen);

	USEMEM(state, tuplen);
	/* reconstruct the HeapTupleData portion */
	tuple->t_len = state->out_store[idx].len ;
	ItemPointerSetInvalid(&(tuple->t_self));
//	tuple->t_datamcxt = CurrentMemoryContext;
	tuple->t_data = (HeapTupleHeader) (((char *) tuple) + HEAPTUPLESIZE);
	/* seek to the tuple */
	if(BufFileSeek(state->myfile,
			state->out_store[idx].fileno,
			state->out_store[idx].offs,SEEK_SET)!=0)
		elog(ERROR,"tupleconn: can't seek in readtup_heap");

	/* read in the tuple proper */
	if (BufFileRead(state->myfile, (void *) tuple->t_data,
					tuple->t_len) != (size_t) tuple->t_len)
		elog(ERROR, "tupleconn: unexpected end of data");

	state->out_store[idx].tup = tuple;
	return (void *) tuple;
}

/*
 * Select comparation function
 * made from SelectSortFunction() in nodeSort.c.
 * differences is the name:) and strategy of operator
 * being searched (here is BTEqualStrategyNumber)
 */
void
SelectConnFunction(Oid connOperator,
				   RegProcedure *connFunction,
				   SortFunctionKind *kind)
{

	CatCList   *catlist;
	int			i;
	HeapTuple	tuple;
	Form_pg_operator optup;
	Oid			opclass = InvalidOid;

	/*
	 * Search pg_amop to see if the target operator is registered as the
	 * "<" or ">" operator of any btree opclass.  It's possible that it
	 * might be registered both ways (eg, if someone were to build a
	 * "reverse sort" opclass for some reason); prefer the "<" case if so.
	 * If the operator is registered the same way in multiple opclasses,
	 * assume we can use the associated comparator function from any one.
	 */
	catlist = SearchSysCacheList(AMOPOPID, 1,
								 ObjectIdGetDatum(connOperator),
								 0, 0, 0);

	for (i = 0; i < catlist->n_members; i++)
	{
		Form_pg_amop aform;

		tuple = &catlist->members[i]->tuple;
		aform = (Form_pg_amop) GETSTRUCT(tuple);

		/* must be of default subtype, too */
		if (aform->amopsubtype != InvalidOid)
			continue;

		if (aform->amopstrategy == BTEqualStrategyNumber)
		{
			opclass = aform->amopclaid;
			*kind = SORTFUNC_CMP;
			break;				/* done looking */
		}
	}

	ReleaseSysCacheList(catlist);

	if (OidIsValid(opclass))
	{
		/* Found a suitable opclass, get its default comparator function */
		*connFunction = get_opclass_proc(opclass, InvalidOid, BTORDER_PROC);
		Assert(RegProcedureIsValid(*connFunction));
		return;
	}

	/*
	 * Can't find a comparator, so use the operator as-is.  Decide whether
	 * it is forward or reverse sort by looking at its name (grotty, but
	 * this only matters for deciding which end NULLs should get sorted
	 * to).  XXX possibly better idea: see whether its selectivity function
	 * is scalargtcmp?
	 */
	tuple = SearchSysCache(OPEROID,
						   ObjectIdGetDatum(connOperator),
						   0, 0, 0);
	if (!HeapTupleIsValid(tuple))
		elog(ERROR, "cache lookup failed for operator %u", connOperator);
	optup = (Form_pg_operator) GETSTRUCT(tuple);
	if (strcmp(NameStr(optup->oprname), "=") == 0)
		*kind = SORTFUNC_CMP;
	else
		*kind = SORTFUNC_LT;
	*connFunction = optup->oprcode;
	ReleaseSysCache(tuple);

	Assert(RegProcedureIsValid(*connFunction));
}


/* ----------------------------------------------------------------
 *		hashFunc
 *
 *  taken from executor/nodeHash.c, it's simplier to copy it here,
 *  than link to nodeHash.c
 * ----------------------------------------------------------------
 */
static uint32
hashFunc(Datum key, int typLen, bool byVal)
{
	unsigned char *k;

	if (byVal)
	{
		/*
		 * If it's a by-value data type, just hash the whole Datum value.
		 * This assumes that datatypes narrower than Datum are
		 * consistently padded (either zero-extended or sign-extended, but
		 * not random bits) to fill Datum; see the XXXGetDatum macros in
		 * postgres.h. NOTE: it would not work to do hash_any(&key, len)
		 * since this would get the wrong bytes on a big-endian machine.
		 */
		k = (unsigned char *) &key;
		typLen = sizeof(Datum);
	}
	else
	{
		if (typLen > 0)
		{
			/* fixed-width pass-by-reference type */
			k = (unsigned char *) DatumGetPointer(key);
		}
		else if (typLen == -1)
		{
			/*
			 * It's a varlena type, so 'key' points to a "struct varlena".
			 * NOTE: VARSIZE returns the "real" data length plus the
			 * sizeof the "vl_len" attribute of varlena (the length
			 * information). 'key' points to the beginning of the varlena
			 * struct, so we have to use "VARDATA" to find the beginning
			 * of the "real" data.	Also, we have to be careful to detoast
			 * the datum if it's toasted.  (We don't worry about freeing
			 * the detoasted copy; that happens for free when the
			 * per-tuple memory context is reset in ExecHashGetBucket.)
			 */
			struct varlena *vkey = PG_DETOAST_DATUM(key);

			typLen = VARSIZE(vkey) - VARHDRSZ;
			k = (unsigned char *) VARDATA(vkey);
		}
		else if (typLen == -2)
		{
			/* It's a null-terminated C string */
			typLen = strlen(DatumGetCString(key)) + 1;
			k = (unsigned char *) DatumGetPointer(key);
		}
		else
		{
			elog(ERROR, "hashFunc: Invalid typLen %d", typLen);
			k = NULL;			/* keep compiler quiet */
		}
	}

	return DatumGetUInt32(hash_any(k, typLen));
}
